//
// Copyright (c) 2014, Peter A.M. van Hoof.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// Peter A.M. van Hoof
// Royal Observatory of Belgium
// Ringlaan 3
// B-1180 Brussels
// Belgium
// p.vanhoof@oma.be
//
//
// If you use any of these data in a scientific publication, please refer to
//
// van Hoof P.A.M., Williams R.J.R., Volk K., Chatzikos M., Ferland G.J., Lykins M., Porter R.L.
// Accurate determination of the free-free Gaunt factor, II -- relativistic Gaunt factors
// 2015, MNRAS, 449, 2112
//

#include "mpreal.h"
#include <vector>
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <string.h>

using namespace std;
using namespace mpfr;

static const mp_prec_t prec_start = 128;

static const long gaunt_magic = 20140210L;
static const long gaunt_magic2 = 20140510L;

#define NP_GAM2 81
#define NP_U 146

double gff[NP_U][NP_GAM2];

double lg_gam2_min, lg_u_min, step, lg_gam2_max, lg_u_max;

inline mpreal operator "" _mp( unsigned long long l )
{
	return mpreal(size_t(l));
}
inline mpreal operator "" _mp( long double l )
{
	return mpreal(l);
}
inline mpreal operator "" _mp( const char* s, size_t )
{
	return mpreal(s);
}

class loss_of_precision
{
public:
	loss_of_precision() {}
};

void gaunt_step2(const mpreal&,const mpreal&,mpreal(*)(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&),
		 double&,double&,const double&,double&,double&,int&);
mpreal g_fun2(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&);

template<typename T>
T pow2(T x) { return x*x; }

size_t min(size_t x, size_t y)
{
	return ( x < y ) ? x : y;
}

size_t max(size_t x, size_t y)
{
	return ( x > y ) ? x : y;
}

void read_table(const char* fnam)
{
	char* ptr;
	long magic;
	size_t np_gam2, np_u, ipu, ipg2, len;
	FILE* io = fopen( fnam, "r" );
	if( io == NULL )
	{
		fprintf( stderr, "failed to open file %s\n", fnam );
		exit(1);
	}
	ptr = NULL;
	/* read magic number */
	while( true )
	{
		getline( &ptr, &len, io );
		if( ptr[0] != '#' )
			break;
		free( ptr );
		ptr = NULL;
	}
	sscanf( ptr, "%ld", &magic );
	free( ptr );
	ptr = NULL;
	if( magic != gaunt_magic && magic != gaunt_magic2 )
	{
		fprintf( stderr, "read_table() found wrong magic number in file %s.\n", fnam );
		exit(1);
	}
	/* read dimensions of the table */
	getline( &ptr, &len, io );
	sscanf( ptr, "%ld %ld",  &np_gam2, &np_u );
	free( ptr );
	ptr = NULL;
	assert( np_gam2 == NP_GAM2 && np_u == NP_U );
	/* read start value for log(gamma^2) */
	getline( &ptr, &len, io );
	sscanf( ptr, "%le",  &lg_gam2_min );
	free( ptr );
	ptr = NULL;
	/* read start value for log(u) */
	getline( &ptr, &len, io );
	sscanf( ptr, "%le",  &lg_u_min );
	free( ptr );
	ptr = NULL;
	/* read step size in dex */
	getline( &ptr, &len, io );
	sscanf( ptr, "%le",  &step );
	free( ptr );
	ptr = NULL;
	lg_gam2_max = lg_gam2_min + (double)(np_gam2-1)*step;
	lg_u_max = lg_u_min + (double)(np_u-1)*step;
	/* read atomic number when present */
	if( magic == gaunt_magic2 )
	{
		getline( &ptr, &len, io );
		free( ptr );
		ptr = NULL;
	}

	/* next lines are comments */
	for( int i=0; i < 9; ++i )
	{
		getline( &ptr, &len, io );
		free( ptr );
		ptr = NULL;
	}

	for( ipu=0; ipu < np_u; ++ipu )
	{
		for( ipg2=0; ipg2 < np_gam2; ++ipg2 )
			fscanf( io, "%le", &gff[ipu][ipg2] );
	}

	/* the remainder of the file contains the uncertainties of the gff values
	 * we will not read those here... */

	fclose(io);
}

double lagrange(const double x[], /* x[n] */
		const double y[], /* y[n] */
		long n,
		double xval)
{
	long i, j;
	double yval = 0.;

	for( i=0; i < n; i++ )
	{
		double l = 1.;
		for( j=0; j < n; j++ )
		{
			if( i != j )
				l *= (xval-x[j])/(x[i]-x[j]);
		}
		yval += y[i]*l;
	}
	return yval;
}

double gauntff(double gam2, double u)
{
	/* use 3rd-order interpolation */
	static const size_t NINTERP = 4;

	size_t i;
	long ipg2, ipu;
	double lg_gam2, lg_u;
	double x[NINTERP], interp[NINTERP];

	assert( gam2 > 0. && u > 0. );

	lg_gam2 = log10(gam2);
	lg_u = log10(u);

	assert( lg_gam2_min <= lg_gam2 && lg_gam2 <= lg_gam2_max );
	assert( lg_u_min <= lg_u && lg_u <= lg_u_max );

	ipg2 = min(max((size_t)((lg_gam2-lg_gam2_min)/step),1)-1,NP_GAM2-NINTERP);
	ipu = min(max((size_t)((lg_u-lg_u_min)/step),1)-1,NP_U-NINTERP);

	for( i=0; i < NINTERP; ++i )
		x[i] = lg_gam2_min + (double)(ipg2+i)*step;

	for( i=0; i < NINTERP; ++i )
		interp[i] = lagrange(x, &gff[ipu+i][ipg2], NINTERP, lg_gam2);

	for( i=0; i < NINTERP; ++i )
		x[i] = lg_u_min + (double)(ipu+i)*step;

	return lagrange(x, interp, NINTERP, lg_u);
}

void gaunt_step2(const mpreal& gam2,
		 const mpreal& u,
		 mpreal (*fun)(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&),
		 double& x,
		 double& h,
		 const double& toler,
		 double& yval,
		 double& dyval,
		 int& error)
{
	if( error > 0 )
		return;

	const int MAX_ITER = 8;
	const int MAX_DIVS = 4;
	const int MAX_ABS = (1<<MAX_DIVS) + 1;

	vector<double> y1(MAX_DIVS+1);
	vector<double> dy1(MAX_DIVS+1);
	vector<double> y1a(MAX_ABS);
	vector<double> y1b(MAX_ABS);

	for( int k=0; k < MAX_ITER; ++k )
	{
		for( int i=0; i <= MAX_DIVS; ++i )
		{
			int nabs = 1<<i;
			int stride = 1<<(MAX_DIVS-i);
			mpreal h1 = mpreal(h)/mpreal(nabs);
			// for k > 0 the points at 0, MAX_DIVS/2 and MAX_DIVS are already set up...
			if( k == 0 || i > 1 )
			{
				for( int j=1; j <= nabs; ++j )
				{
					int index = j*stride;
					// y1a[index] may already have been initialized on a previous
					// iteration of the loop over i. If so, do not repeat the calculation
					if( i == 0 || j%2 == 1 )
					{
						mpreal x1 = x + h1*mpreal(j);
						mpreal rerr;
						mp_prec_t myprec;
						y1a[index] = double((*fun)(x1,gam2,u,rerr,myprec));
						y1b[index] = double(rerr*y1a[index]);
					}
				}
			}
			if( i < 2 ) {
				y1[i] = 0.;
				dy1[i] = 0.;
			}
			else {
				// use Eq. 4.1.20 from Numerical Recipes in C, 1st ed.
				y1[i] = (23.*y1a[stride]+7.*y1a[2*stride]+
					 13.*y1a[(nabs-1)*stride]+5.*y1a[nabs*stride])/12.;
				for( int j=3; j < nabs-1; ++j )
					y1[i] += y1a[j*stride];
				y1[i] *= double(h1);
				// this is the regular error propagation formula assuming errors are uncorrelated
				dy1[i] = (pow2(23.*y1b[stride])+pow2(7.*y1b[2*stride])+
					  pow2(13.*y1b[(nabs-1)*stride])+pow2(5.*y1b[nabs*stride]))/144.;
				for( int j=3; j < nabs-1; ++j )
					dy1[i] += pow2(y1b[j*stride]);
				dy1[i] = sqrt(dy1[i])*double(h1);
			}
			// check for convergence by comparing the integrals with stepsizes h1 and 2*h1
			if( i > 2 )
			{
				// toler is relative precision
				double err = abs((y1[i]-y1[i-1])/((yval+y1[i])*toler));
				if( err <= 1. )
				{
					double err = hypot(dy1[i],abs(y1[i]-y1[i-1])/2.);
					yval += y1[i];
					dyval = hypot(dyval,err);
					double fac = 0.9*cbrt(1./max(err,0.1));
					x += h;
					h = 8.*h1*fac;
					// make estimate of the remainder of the integral to infinity
					// assuming that g_iii() is constant
					// use additional factor 100 to make sure that the error from
					// truncating the integral is negligible compared to the other
					// terms and our error estimate is valid
					error = ( abs(y1[i]/(yval*expm1(h))) < toler/100. ) ? 1 : 0;
					return;
				}
			}
		}
		// convergence failed. Here we reduce the stepsize and try again...
		y1a[(MAX_ABS-1)/2] = y1a[1];
		y1a[MAX_ABS-1] = y1a[2];
		y1b[(MAX_ABS-1)/2] = y1b[1];
		y1b[MAX_ABS-1] = y1b[2];
		h /= double((MAX_ABS-1)/2);
	}
	error = 2;
	return;
}

mpreal g_fun2(const mpreal& x, const mpreal& gam2, const mpreal&, mpreal& rerr, mp_prec_t& myprec)
{
	mpreal err;
	mpreal res = gauntff(gam2,x);
	rerr = err/res;
	myprec = 0;
	return exp(-x)*res;
}

double g_iii_fr_av(const mpreal& gam2, double& err, const double& tol)
{
	mpreal u = 0_mp;
	double x = 0.;
	double h = max(sqrt(double(gam2)/15.), 1.)*min(1.e-5, tol);
	double res = 1.;
	err = 0.;
	// error == 0 : integration needs more steps
	// error == 1 : normal termination of integral
	// error == 2 : step failed to converge
	int error = 0;
	do {
		gaunt_step2( gam2, u, g_fun2, x, h, tol/2., res, err, error );
	} while( error == 0 );

	if( error < 2 )
		return res - 1.;
	else {
		cerr << "integration failed: gam2 = " << gam2 << " u = " << u << endl;
		throw loss_of_precision();
	}
}

int main(int, char** argv)
{
	mpreal::set_default_prec(prec_start);

	string fnam = argv[1];
	read_table( fnam.c_str() );

	unsigned int g2max = 161;

	mpreal g2start = -6_mp;
	mpreal steps_per_dex = 10_mp;

	vector<double> resv( g2max );
	vector<double> dresv( g2max );
	for( unsigned int i=0; i < resv.size(); ++i ) {
		resv[i] = 0.;
		dresv[i] = 0.;
	}

	for( unsigned int g2=0; g2 < g2max; ++g2 ) {
		mpreal gam2 = mpreal(g2)/steps_per_dex + g2start;
		double err, res;
		res = g_iii_fr_av( exp10(gam2), err, 1.e-6 );
		resv[ g2 ] = res;
		dresv[ g2 ] = err;
	}

	auto p = fnam.find( 'Z' );
	string Zstr = fnam.substr( ++p, 2 );
	
	cout << setw(9) << 20141008 << " # magic number" << endl;
	cout << setw(9) << g2max << " # number of grid points in gam2 direction" << endl;
	cout << setw(9) << g2start << " # start value for log(gam2)" << endl;
	cout << setw(9) << 1_mp/steps_per_dex << " # stepsize in dex for log(gam2)" << endl;
	cout << setw(9) << Zstr << " # atomic number Z" << endl;
	cout << "#" << endl;
	cout << "# log(gam^2)  <g_ff(gam^2)>" << endl;
	for( unsigned int k=0; k < resv.size(); ++k ) {
		cout << fixed << setprecision(2) << setw(6) << mpreal(k)/steps_per_dex + g2start;
		cout << scientific << setprecision(5) << setw(13) << resv[k] << endl;
	}

	return 0;
}
